/***************************************************************************
 *   Copyright (C) 2007 by Artur Szymiec   *
 *   artur.szymiec@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include <stdlib.h>
#include <stdarg.h>
#include <qcheckbox.h>
#include <qcolor.h>
#include <qslider.h>
#include <kapplication.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <kpushbutton.h>
#include <kconfig.h>
#include <kglobal.h>
#include <kstandarddirs.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <png.h>
#include "kmatrix3d.h"
#include "kmatrix3dui.h"

//! libkscreensaver interface
extern "C"
{
        const char *kss_applicationName = "kmatrix3d.kss";
        const char *kss_description = I18N_NOOP( "KMatrix3D" );
        const char *kss_version = "2.2.0";

        KMatrix3D *kss_create( WId id ) {
                KGlobal::locale()->insertCatalogue("kmatrix3d");
                return new KMatrix3D( id );
        }

        QDialog *kss_setup() {
                KGlobal::locale()->insertCatalogue("kmatrix3d");
                return new KMatrix3DSetup();
        }
}

//-----------------------------------------------------------------------------
//! dialog to setup screen saver parameters
KMatrix3DSetup::KMatrix3DSetup( QWidget *parent, const char *name )
                : KMatrix3DUI( parent, name, TRUE )
{
        //Connect your signals and slots here to configure the screen saver.
       connect( OkayPushButton, SIGNAL( released() ),
                SLOT( slotOkPressed() ) );
        connect( CancelPushButton, SIGNAL( released() ),
                SLOT( slotCancelPressed() ) );
	//Read Settings
	readSettings();
}


//! read settings from config file
void KMatrix3DSetup::readSettings()
{
        KConfig *config = KGlobal::config();
        config->setGroup( "Settings" );
	// config options here...
	sld3dlines->setValue(config->readNumEntry("NoOf3DLines",800));
}


//! Ok pressed - save settings and exit
void KMatrix3DSetup::slotOkPressed()
{
        KConfig *config = KGlobal::config();
        config->setGroup( "Settings" );
        // config options here.
        config->writeEntry("NoOf3DLines",sld3dlines->value());
	config->sync();

        accept();
}

void KMatrix3DSetup::slotCancelPressed()
{
        reject();
}
//-----------------------------------------------------------------------------


KMatrix3D::KMatrix3D( WId id ) : KScreenSaver( id )
{
        readSettings();
        blank();
}

KMatrix3D::~KMatrix3D()
{}


//! read configuration settings from config file
void KMatrix3D::readSettings()
{
        KConfig *config = KGlobal::config();
        config->setGroup( "Settings" );
        // config options here...
        config3dLines = config->readNumEntry("NoOf3DLines",800);
}


void KMatrix3D::blank()
{
        
        //code to render the screen.
	if(config3dLines<50)
		config3dLines=50;
	if(config3dLines>NBR_LIGNES_MAX_DEF)
		config3dLines=NBR_LIGNES_MAX_DEF;
        this->myGL = new mQGL(config3dLines);
        embed(myGL);
        myGL->show();
}

// Our mQGL member functions
mQGL::mQGL(int noOf3DLines)
{
        //set new camera
	NBR_LIGNES_MAX=noOf3DLines;
        camera_1=new camera( 0 , 0 , 0 , 0 , 0 , 0 , 175 , &s_infos_gen );
        // space for sorting FIXME delete sorting at all
        s_infos_gen.ordre_sz = new ordre_z[NBR_LIGNES_MAX];
        //set new lines
        for (int i=0 ; i<NBR_LIGNES_MAX ; i++) {

                ligne_1[i] = new Ligne3D1( double((rand()%200)-100) , double((rand()%200)-100) , double((rand()%200)-100) , &s_infos_gen , i);

        }
        this->i=0;
        this->i2=0;
}

void mQGL::initializeGL()
{
        //init OpenGL
        glEnable(GL_TEXTURE_2D);
	// This should be enabled while using GL_ONE_MINUS_DST_ALPHA
	// but seems that some bug exist in KDE and mode currently used is
	// GL_ONE - fake trancparency but with sorting work ok
        glDisable(GL_DEPTH_TEST); 
        //glClearColor( 0.0, 0.0, 0.0, 1.0 );
        // set mode to blending
        glEnable(GL_BLEND);
        // set blend mode
        //glBlendFunc( GL_SRC_ALPHA,GL_ONE_MINUS_DST_ALPHA);
        glBlendFunc(GL_SRC_ALPHA,GL_ONE);
        //set shading model
        glShadeModel(GL_SMOOTH);
        // set fl correction of perspective
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

        //init textures
        if(!InitTextures())
                abort();
}

int mQGL::InitTextures()
{
        int i=0 , i2=1 , i3=0;
 	KStandardDirs *kdedirs;
	kdedirs = new KStandardDirs();
	QString str1,str2,str3;	

	// Locate textures
	str1=kdedirs->findResourceDir("data", "kmatrix3d/textures/fontes_vertes_alpha.png");
	str2=kdedirs->findResourceDir("data", "kmatrix3d/textures/fontes_blanches_alpha.png");
	str3=kdedirs->findResourceDir("data", "kmatrix3d/textures/nuage1.png");
	str1.append("kmatrix3d/textures/fontes_vertes_alpha.png");
	str2.append("kmatrix3d/textures/fontes_blanches_alpha.png");
	str3.append("kmatrix3d/textures/nuage1.png");
	
        // load the textures
	//"$kde_datadir/fontes_vertes_alpha.png
        if(!LoadTexture( s_infos_gen.tex_lettres_vertes , (char*) str1.latin1() ))
                return false;
        if(!LoadTexture( s_infos_gen.tex_lettres_blanches , (char*) str2.latin1() ))
                return false;
        if(!LoadTexture( s_infos_gen.tex_part_1 , (char*) str3.latin1() ))
                return false;
        // Initialize the list of texture cordinates
        // for vertex of characters 0 to 56
        for(i=0 ; i<56 ; i++) {
                // calculate the cordinates for particular letter
                s_infos_gen.info_tex_vertes_sz[i].x1 = float(0.0976 * i3);
                s_infos_gen.info_tex_vertes_sz[i].y1 = float(0.113 * i2);

                s_infos_gen.info_tex_vertes_sz[i].x2 = float(0.0976 * i3);
                s_infos_gen.info_tex_vertes_sz[i].y2 = float(0.113 * (i2-1));

                s_infos_gen.info_tex_vertes_sz[i].x3 = float(0.0976 * (i3+1));
                s_infos_gen.info_tex_vertes_sz[i].y3 = float(0.113 * (i2-1));

                s_infos_gen.info_tex_vertes_sz[i].x4 = float(0.0976 * (i3+1));
                s_infos_gen.info_tex_vertes_sz[i].y4 = float(0.113 * i2);

                // increment i3 - the counter of chars in row
                i3++;

                // check that if i3 more then 9
                if( i3 > 9 ) {
                        // if so reset the i3 to 0
                        i3 = 0;

                        // and add 1 to i2 - a row counter
                        i2++;

                }

        }
        return true;
}

int mQGL::LoadTexture(unsigned &texture_sz,char * filename)
{
        int width, height;
        png_byte color_type;
        png_byte bit_depth;

        png_structp png_ptr;
        png_infop info_ptr;
        int number_of_passes;
        png_bytep * row_pointers;
        int y;
        void *mybuffer;
        char header[8];	// 8 is the maximum size that can be checked

        //OpenGL
        GLenum *gluerr=new GLenum;
        /* open file and test for it being a png */
        FILE *fp = fopen(filename, "rb");
        if (!fp)
                abort_("[read_png_file] File %s could not be opened for reading", filename);
        fread(header, 1, 8, fp);
        if (png_sig_cmp((png_byte*)header, 0, 8))
                abort_("[read_png_file] File %s is not recognized as a PNG file", filename);


        /* initialize stuff */
        png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

        if (!png_ptr)
                abort_("[read_png_file] png_create_read_struct failed");

        info_ptr = png_create_info_struct(png_ptr);
        if (!info_ptr)
                abort_("[read_png_file] png_create_info_struct failed");

        if (setjmp(png_jmpbuf(png_ptr)))
                abort_("[read_png_file] Error during init_io");

        png_init_io(png_ptr, fp);
        png_set_sig_bytes(png_ptr, 8);

        png_read_info(png_ptr, info_ptr);

        width = info_ptr->width;
        height = info_ptr->height;
        color_type = info_ptr->color_type;
        bit_depth = info_ptr->bit_depth;

        number_of_passes = png_set_interlace_handling(png_ptr);
        png_set_gamma(png_ptr, 1.2, 1.0/2.2); //FIXME decide something about it
        png_read_update_info(png_ptr, info_ptr);


        /* read file */
        if (setjmp(png_jmpbuf(png_ptr)))
                abort_("[read_png_file] Error during read_image");

        mybuffer=malloc(sizeof(png_bytep) * info_ptr->rowbytes * height);
        row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * height);
        for (y=0; y<height; y++)
                //row_pointers[y] = (png_byte*) malloc(info_ptr->rowbytes);
                row_pointers[y]=(png_byte*) mybuffer+info_ptr->rowbytes*y;
        png_read_image(png_ptr, row_pointers);

        fclose(fp);
        //OpenGL
        glGenTextures(1,&texture_sz);
        glBindTexture(GL_TEXTURE_2D,texture_sz);

        //Generate texture mipmaps
        if (gluBuild2DMipmaps(GL_TEXTURE_2D, 4, width, height,
                              GL_RGBA, GL_UNSIGNED_BYTE,mybuffer)) {

                fprintf(stderr,"GLULib%s\n",gluErrorString(*gluerr));
                exit(-1);
        }
        // Some pretty standard settings for wrapping and filtering.
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
        glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR_MIPMAP_LINEAR);
        //glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
        //free resources
        free(row_pointers);
        free(mybuffer);
        return true;
}

void mQGL::timerEvent( QTimerEvent * )
{
        updateGL();
}

void mQGL::abort_(const char * s, ...)
{
        KMessageBox kmsg;
        va_list args;
        va_start(args, s);
        vfprintf(stderr, s, args);
        fprintf(stderr, "\n");
        kmsg.error(0,"Error reading textures","Error",0);
        va_end(args);
        abort();
}

void mQGL::resizeGL(int width, int height)
{
        glViewport(0,0,width,height);

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        gluPerspective (45,(double)width/(double)height,1,100000);

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

}

void mQGL::paintGL()
{
        // initialize our timer temps_1 with current time value
        temps_1 = clock();
        // calculate the msec time passed already from last update
        s_infos_gen.msec = (int)(1000*(temps_1 - temps_2)/(float) CLOCKS_PER_SEC);

        // clear screen
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // load identity matrix
        glLoadIdentity();

        // calculate the number of images per second
//         if((float)(1000*(clock()- fps_2))/(float) CLOCKS_PER_SEC >= 1000) {
//                 gettimeofday(&r_time1,NULL);
//                 timeval_subtract(&r_time3,&r_time1,&r_time2);
//                 fps_img_2 = fps_img_1;
//                 fps_real=(float)fps_img_2/((float)r_time3.tv_sec+(float)r_time3.tv_usec/1000000);
//                 //printf("fps now = %.1f\n",fps_real);
//                 fps_img_1 = 0;
//                 gettimeofday(&r_time2,NULL);
//                 fps_2=clock();
// 		
//         }

        // increase the number of generated pictures / sec
        fps_img_1++;


        // calculate position of the camera
        camera_1->calculer();

        // calculate particular line
        for (i=0 ; i<NBR_LIGNES_MAX ; i++) {

                ligne_1[i]->calculer();

        }
	//Sort iz Z order 
        heapsort(NBR_LIGNES_MAX,&s_infos_gen.ordre_sz[0]);
        // and the display function of position to the camera (orientation + or -)
        if((s_infos_gen.infos_cam_sz.m_z_v - s_infos_gen.infos_cam_sz.m_z_t) < 0) {

                for (i=NBR_LIGNES_MAX-1 ; i>=0 ; i--) {
                        ligne_1[s_infos_gen.ordre_sz[i].index]->afficher();
                }

        } else {


                for (i=0 ; i<NBR_LIGNES_MAX ; i++) {

                        ligne_1[s_infos_gen.ordre_sz[i].index]->afficher();
                }

        }
        // All done drawing.  Let's show it.
        //glutSwapBuffers(); //Must use swapBuffers of qglWidget
        //swapBuffers(); //Not necessary - autoswaping on by default
	//enable timer for next refresh
	startTimer(1);
        //set the timer (for fps calc)
        temps_2 = temps_1;
}

int mQGL::timeval_subtract (struct timeval *result,struct timeval *x,struct timeval *y)
{
        /* Perform the carry for the later subtraction by updating y. */
        if (x->tv_usec < y->tv_usec)
        {
                int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
                y->tv_usec -= 1000000 * nsec;
                y->tv_sec += nsec;
        }
        if (x->tv_usec - y->tv_usec > 1000000)
        {
                int nsec = (x->tv_usec - y->tv_usec) / 1000000;
                y->tv_usec += 1000000 * nsec;
                y->tv_sec -= nsec;
        }

        /* Compute the time remaining to wait.
           tv_usec is certainly positive. */
        result->tv_sec = x->tv_sec - y->tv_sec;
        result->tv_usec = x->tv_usec - y->tv_usec;

        /* Return 1 if result is negative. */
        return x->tv_sec < y->tv_sec;
}

//Heap sorting part 1
void mQGL::sift_in(unsigned count, ordre_z *data, unsigned free, ordre_z next)
{

        unsigned  i;
        //** sift up the free node **//
        for (i=2*free;i<count;i+=i) {
                if (data[i].ordre < (data[i+1].ordre))
                        i++;
                {
                        data[free]=data[i];
                        free=i;
                }
        }
        /* special case in sift up if the last inner node has only 1 child */
        if (i==count) {
                data[free]=data[i];
                free=i;
        }
        //** sift down the new item next **//
        while((i=free/2)  &&  (data[i].ordre < next.ordre)) {
                data[free]=data[i];
                free=i;
        }
        data[free]= next;
}

//Heap sorting part 2
void mQGL::heapsort(unsigned count, ordre_z *data)
{
        unsigned  j;

        if (count <= 1)
                return;
        data-=1;   /* map addresses to indices 1 til count  */
        //** build the heap structure **//
        for(j=count/2;j;j-=1) {
                ordre_z  next = data[j];
                sift_in(count, data, j, next);
        }
        //** search next by next remaining extremal element **//
        for(j=count-1;j;j-=1) {
                ordre_z  next = data[j+1];
                data[j+1] = data[1];   /* extract extremal element from the heap */
                sift_in(j, data, 1, next);
        }

}





