/***************************************************************************
 *   Copyright (C) 2007 by Artur Szymiec                                   *
 *   artur.szymiec@gmail.com                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *                      Orginal Code & Idea by                             *
 *                          Thomas John                                    *
 *                      thomas.john@open-design.be                         *
 *                                                                         *
 *                Linux Port & Some Minor Changes by                       *
 *                         Artur Szymiec                                   *
 *                   artur.szymiec@gmail.com	                           *
 ***************************************************************************/

#include "ligne3d1.h"
#include <GL/gl.h>
#include <math.h>
#include <stdio.h>
#include <cstdlib>

// initialization of the letter
void Ligne3D1::init_lettre( int index_sz )
{

        // initialise all variables of this letter
        // it is important initializer when class is created
        // before utilization

        // cordinates
        lettre_1[index_sz].pos_x = m_x;
        lettre_1[index_sz].pos_y = m_y - (0.7 * m_nbr_actif); // every active letter * 0.7
        lettre_1[index_sz].pos_z = m_z;

        // tat de la lettre 1 = dbut
        // state of the letter 1 = begining
        lettre_1[index_sz].etat = 1;

        // timer 1
        lettre_1[index_sz].temps = 0;
        lettre_1[index_sz].temps_tot = m_temps_tot;

        // character number out of 56 that we're going to display
        lettre_1[index_sz].caractere = rand()%55;

        // opacity of the character
        lettre_1[index_sz].opacite = 100;

        // distance from camera
        lettre_1[index_sz].dist_tmp = 0;

        // specify time when the letter can change
        lettre_1[index_sz].change_sz = rand() % 3;

        // timer 2 (change of the character)
        lettre_1[index_sz].temps_2 = 0;
        lettre_1[index_sz].temps_tot_2 = 70 + (rand() % 500);

}

// default constructor
Ligne3D1::Ligne3D1( double x, double y, double z , struct infos_gen * infos_sz , int index)
{
        // init global variables

        // pointer to the structure with global info
        infos_gen_tmp = infos_sz;

        // cordinates
        m_y = double((rand()%100)-50) + infos_gen_tmp->infos_cam_sz.m_y_v;
        m_z = double((rand()%100)) + infos_gen_tmp->infos_cam_sz.m_z_v;
        // plus the distance of camera in x vector (should be bigger)
        m_x = double((rand()%100)-50) + infos_gen_tmp->infos_cam_sz.m_x_v;
        /*
        	// sets for test
        	m_x = infos_gen_tmp->infos_cam_sz.m_x_t;
        	m_y = infos_gen_tmp->infos_cam_sz.m_y_t;
        	m_z = infos_gen_tmp->infos_cam_sz.m_z_t;
        */
        // state of the instance (1 = active)
        m_etat = 1;

        // max number of letters displayed
        nbr_max_lettres = 20 + int(rand()%100);

        // set counting down
        decpt_nbr_lettres = nbr_max_lettres;

        // on create our table with information about each and every letter
        lettre_1 = new Lettre3D1[nbr_max_lettres+1];

        // number of letters active
        m_nbr_actif = 0;

        // timer 1
        m_temps = 0;
        m_temps_tot = 1 + (rand()%300);

        // index of the instance
        index_tmp = index;

        // duration of melting white->green
        m_temps_fondu = m_temps_tot;

        // set the flag instance of particles charged to false
        m_charge_part = false;

        // if speed speed of the lettres is not rapid (<30)
        // and the number of letters displayed >60 display particles
        if(m_temps_tot<30 && nbr_max_lettres>60)
        {
                // set flag instance of particles to TRUE
                m_effet_particule = true;

                // create the instance of particles
                particules_1 = new effet_particule( 0 , 0 , 0 , 1 , 200 , infos_sz , 1 , 1);
                particules_1->m_x = m_x;
                particules_1->m_y = m_y;
                particules_1->m_z = m_z;

                // one instance of the class was charged
                m_charge_part = true;

        } else
        {

                // otherwise set particle effect flag to FALSE
                m_effet_particule = false;

        }

        // Distance line to camera
        dist_tmp = 0;

}

// destructor
Ligne3D1::~Ligne3D1( )
{

        // od desruction and if exist a particle effect delete it as well
        if(m_charge_part == true)
                delete particules_1;

}

// calculat the position and state of every element of the line
void Ligne3D1::calculer()
{
        // check if we can create new letter
        if(m_nbr_actif < nbr_max_lettres) {

                // increment our timer
                m_temps += infos_gen_tmp->msec;

                // check that our timer not pass the end
                if (m_temps >= m_temps_tot) {

                        // increase the number of active letters
                        m_nbr_actif++;

                        // initialise new letter
                        init_lettre(m_nbr_actif);

                        // re-initialize our timer
                        m_temps = 0;

                }

        }

        // calculate the state of the letters
        for ( i = 1 ; i <= m_nbr_actif ; i++) {

                // check if the letter is still active
                if(lettre_1[i].etat > 0) {

                        // increment the timer
                        lettre_1[i].temps += infos_gen_tmp->msec;

                        // ok, in fact this is the function of the state for each letter
                        switch(lettre_1[i].etat) {

                        case 1:	// done with white letter

                                // calculate the percentage
                                prct_sz = (double(lettre_1[i].temps) / double(lettre_1[i].temps_tot));

                                // change opacity
                                lettre_1[i].opacite = prct_sz;
                                //lettre_1[i].opacite = prct_sz * opacite_gen; // can be used

                                break;

                        case 2: // display white letter

                                // change opacity to 1
                                lettre_1[i].opacite = 1.0;
                                //lettre_1[i].opacite = opacite_gen; // can be used

                                break;

                        case 3: // transition white->green

                                // calculate the percentage
                                prct_sz = (double(lettre_1[i].temps) / double(lettre_1[i].temps_tot));

                                // change opacity
                                lettre_1[i].opacite = prct_sz;
                                //lettre_1[i].opacite = prct_sz * opacite_gen; // can be used

                                break;

                        case 4:	// display green letter

                                // change opacity
                                lettre_1[i].opacite = 1.0;
                                //lettre_1[i].opacite = opacite_gen; // can be used

                                break;

                        case 5: // change down opacity of the letter

                                // calculate percentage
                                prct_sz = 1.0 - (double(lettre_1[i].temps) / double(lettre_1[i].temps_tot));

                                // change opacity
                                lettre_1[i].opacite = prct_sz;
                                //lettre_1[i].opacite = prct_sz * opacite_gen; // can be used

                                break;

                        }

                        // check if letter can change
                        if( lettre_1[i].change_sz == 1 ) {

                                // yes, update our timer with respect to that
                                lettre_1[i].temps_2 += infos_gen_tmp->msec;

                                // if the timer arrived to end change the character
                                if ( lettre_1[i].temps_2 >= lettre_1[i].temps_tot_2 ) {

                                        // re-initialize our timer
                                        lettre_1[i].temps_2 = 0;

                                        // and change the character
                                        lettre_1[i].caractere = int(rand()%55);

                                }

                        }

                        // modify the letter opacity in the function of camera - letter distance
                        dist_x = infos_gen_tmp->infos_cam_sz.m_x_v - m_x;
                        dist_y = infos_gen_tmp->infos_cam_sz.m_y_v - m_y;
                        dist_z = infos_gen_tmp->infos_cam_sz.m_z_v - m_z;

                        // calculate distance
                        lettre_1[i].dist_tmp = sqrt( (dist_x * dist_x) + (dist_y * dist_y) + (dist_z * dist_z));


                        // check the the letter is not to far ??
                        if(lettre_1[i].dist_tmp > 200) {

                                // yes so set opacity to 0
                                lettre_1[i].opacite = 0;

                        } else if( lettre_1[i].dist_tmp > 100 ) {

                                // no - but within range 100-200 so change the opacity in the function of distance
                                lettre_1[i].opacite = (1 - ((lettre_1[i].dist_tmp-100) / 100)) * lettre_1[i].opacite;

                        }


                }

        }

        //recover the distance from camera from previous letter
        dist_tmp = lettre_1[i-1].dist_tmp;

        // if line use the "particle effect"
        if(m_effet_particule == true) {

                // if number of active created letters in line reached max
                if(m_nbr_actif >= nbr_max_lettres) {
                        //yes - stopthe emiision of new particles
                        particules_1->etat = 2;

                }

                // put the 'y' cordiante of our emmision particle
                particules_1->m_y = m_y - (0.7 * m_nbr_actif);

                // calculate the effect
                particules_1->calculer();

        }


        // specify cordinates in table which serve sorting all the lines
        // for display far-off
        infos_gen_tmp->ordre_sz[index_tmp].index = index_tmp;
        infos_gen_tmp->ordre_sz[index_tmp].ordre = m_z;

}

// display elements
void Ligne3D1::afficher()
{

        i=0;
	i2=0;
	i3=0;

        // go for each letter
        for ( i = 1 ; i <= m_nbr_actif ; i++) {

                // select function for state of the letter
                switch (lettre_1[i].etat) {

                case 0: // letter not active nothing to do
                        break;

                case 1: // display white letter

                        // verify that the opacity is more then 0
                        if(lettre_1[i].opacite > 0) {

                                // set the GL texture used to display
                                glBindTexture(GL_TEXTURE_2D, infos_gen_tmp->tex_lettres_blanches);

                        }

                        // verify that the leter display time is finished
                        if ( lettre_1[i].temps >= lettre_1[i].temps_tot ) {

                                // change state if the letter
                                lettre_1[i].etat++;

                                // set timer to 0
                                lettre_1[i].temps = 0;

                                // set new duration for our border timer value
                                lettre_1[i].temps_tot = 1;

                        }

                        break;

                case 2: // display white letter

                        if(lettre_1[i].opacite > 0) {

                                // set the GL texture used to display
                                glBindTexture(GL_TEXTURE_2D, infos_gen_tmp->tex_lettres_blanches);

                        }

                        // verify that the leter display time is finished
                        if ( lettre_1[i].temps >= lettre_1[i].temps_tot ) {

                                // change the state of the letter
                                lettre_1[i].etat++;

                                // set timer to 0
                                lettre_1[i].temps = 0;

                                // set new duration for our border timer value white->green
                                lettre_1[i].temps_tot = m_temps_fondu;

                        }

                        break;

                case 3: // melting white -> green


                        // verify that the distance letter-camera is less then 201 (otherwise display nothing)
                        if(lettre_1[i].dist_tmp < 201) {

                                // display green letter
                                // set the GL texture used to display
                                glBindTexture(GL_TEXTURE_2D, infos_gen_tmp->tex_lettres_blanches);


                                // opacity of the letter is difference between 1 and letter distance to camera and opacity
                                // if tge distance is between 100 and 200
                                // reverse necessary for alpha (1-value)
                                if( lettre_1[i].dist_tmp < 200 && lettre_1[i].dist_tmp > 100 ) {

                                        // set the opacity
                                        glColor4d( 1.0 , 1.0 , 1.0 , (1 - ((lettre_1[i].dist_tmp-100) / 100)) - lettre_1[i].opacite);


                                } else {

                                        // set opacity
                                        glColor4d( 1.0 , 1.0 , 1.0 , (1 - lettre_1[i].opacite));

                                }


                                // draw the white letter !
                                glBegin(GL_QUADS);

                                glTexCoord2f( infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].x1 , infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].y1 );
                                glVertex3d( lettre_1[i].pos_x, lettre_1[i].pos_y , lettre_1[i].pos_z);

                                glTexCoord2f( infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].x2 , infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].y2 );
                                glVertex3d( lettre_1[i].pos_x, lettre_1[i].pos_y + 1.0 , lettre_1[i].pos_z);

                                glTexCoord2f( infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].x3 , infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].y3 );
                                glVertex3d( lettre_1[i].pos_x + 1.0, lettre_1[i].pos_y + 1.0 , lettre_1[i].pos_z);

                                glTexCoord2f( infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].x4 , infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].y4 );
                                glVertex3d( lettre_1[i].pos_x + 1.0, lettre_1[i].pos_y , lettre_1[i].pos_z);

                                glEnd();

                                // set the GL texure (this time for green characters)
                                glBindTexture(GL_TEXTURE_2D, infos_gen_tmp->tex_lettres_vertes);
                        }

                        // verify that the leter display time is finished
                        if ( lettre_1[i].temps >= lettre_1[i].temps_tot ) {

                                // change the state of the letter
                                lettre_1[i].etat++;

                                // reset timer value
                                lettre_1[i].temps = 0;

                                // set new duration
                                lettre_1[i].temps_tot = 500 + (rand() % 2000);

                        }

                        break;

                case 4: // display green letter

                        if(lettre_1[i].opacite > 0) {

                                // set the GL texure
                                glBindTexture(GL_TEXTURE_2D, infos_gen_tmp->tex_lettres_vertes);

                        }

                        // verify that the leter display time is finished
                        if ( lettre_1[i].temps >= lettre_1[i].temps_tot ) {

                                // change the state
                                lettre_1[i].etat++;

                                // reset timer value to 0
                                lettre_1[i].temps = 0;

                        }

                        break;

                case 5: // change "green" opacity

                        if(lettre_1[i].opacite > 0) {

                                // set the GL texure
                                glBindTexture(GL_TEXTURE_2D, infos_gen_tmp->tex_lettres_vertes);

                        }

                        //Calculate the Z,X cordinates ration for clipping if behind field of view
                        if(infos_gen_tmp->infos_cam_sz.m_z_v>=m_z) {
                                a1=fabs(infos_gen_tmp->infos_cam_sz.m_z_v-(double)m_z);
                        } else {
                                a1=fabs((double)m_z-infos_gen_tmp->infos_cam_sz.m_z_v);
                        }
                        if(infos_gen_tmp->infos_cam_sz.m_x_v>=m_x) {
                                a2=fabs(infos_gen_tmp->infos_cam_sz.m_x_v-(double)m_x);
                        } else {
                                a2=fabs((double)m_x-infos_gen_tmp->infos_cam_sz.m_x_v);
                        }

                        // verify that the leter display time is finished or behind camera
                        if ( lettre_1[i].temps >= lettre_1[i].temps_tot || (a1/a2<1.7) ) {
                                // change the state to 0 - no active
                                lettre_1[i].etat = 0;

                                // reset timer to 0
                                lettre_1[i].temps = 0;

                                // decrease active letters
                                decpt_nbr_lettres--;

                                // verify that all letters have "gone" to 0 or behind camera
                                if(decpt_nbr_lettres == 0 || a1/a2<1.7) {

                                        // line re-initialize
                                        m_y = double((rand()%150)-50) + infos_gen_tmp->infos_cam_sz.m_y_t;
                                        i3 = (rand()%200)-100;
                                        m_z = i3 + infos_gen_tmp->infos_cam_sz.m_z_t;
                                        // set the x position if the line to compute from camera far enough
                                        m_x = double((rand()%100)-50) + infos_gen_tmp->infos_cam_sz.m_x_t;


                                        // change the state of the instance
                                        m_etat = 1;

                                        // set down counting
                                        decpt_nbr_lettres = nbr_max_lettres;

                                        // set number of active letters
                                        m_nbr_actif = 0;

                                        // reset timer
                                        m_temps = 0;

                                        // set speed of new created letters
                                        m_temps_tot = 1 + int((rand()%300));

                                        // if the speed is not rapid and number of letters is more then 60
                                        // draw particle
                                        if(m_temps_tot<30 && nbr_max_lettres>60) {

                                                // verify that instance of the class particle wasn't created yet
                                                if(m_charge_part == false) {

                                                        // no - charge new particle
                                                        particules_1 = new effet_particule( 0 , 0 , 0 , 1 , 200 , infos_gen_tmp , 1 , 1);

                                                        m_charge_part = true;

                                                }

                                                // set effect variable to TRUE
                                                m_effet_particule = true;

                                                // set cordinates and initialize particle
                                                particules_1->m_x = m_x;
                                                particules_1->m_y = m_y;
                                                particules_1->m_z = m_z;
                                                particules_1->etat = 1;
                                                particules_1->m_nombre_actifs=0;

                                        } else {

                                                // otherwise set particle effect flag to false
                                                m_effet_particule = false;

                                        }

                                }

                        }

                        break;

                }

                // verify that the letter state is active and opacity positive
                if(lettre_1[i].etat > 0 && lettre_1[i].opacite > 0) {

                        // set opacity
                        glColor4d( 1.0 , 1.0 , 1.0 , lettre_1[i].opacite );

                        // draw the letter
                        glBegin(GL_QUADS);

                        glTexCoord2f( infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].x1 , infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].y1 );
                        glVertex3d( lettre_1[i].pos_x, lettre_1[i].pos_y , lettre_1[i].pos_z);

                        glTexCoord2f( infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].x2 , infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].y2 );
                        glVertex3d( lettre_1[i].pos_x, lettre_1[i].pos_y + 1.0 , lettre_1[i].pos_z);

                        glTexCoord2f( infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].x3 , infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].y3 );
                        glVertex3d( lettre_1[i].pos_x + 1.0, lettre_1[i].pos_y + 1.0 , lettre_1[i].pos_z);

                        glTexCoord2f( infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].x4 , infos_gen_tmp->info_tex_vertes_sz[lettre_1[i].caractere].y4 );
                        glVertex3d( lettre_1[i].pos_x + 1.0, lettre_1[i].pos_y , lettre_1[i].pos_z);

                        glEnd();

                }

        }

        // if the line uses the particle effect - and distance is less the 201
        if(m_effet_particule == true && dist_tmp < 201) {

                // change blending mode to display particles
                //glBlendFunc(GL_SRC_ALPHA,GL_ONE);
                glBlendFunc(GL_SRC_ALPHA,GL_DST_COLOR); //FIXME decide something
                //glBlendFunc(GL_ONE,GL_ONE);
                //glBlendFunc(GL_SRC_COLOR,GL_ONE);
                //glBlendFunc(GL_SRC_ALPHA,GL_DST_ALPHA);

                // set GL texture used
                glBindTexture(GL_TEXTURE_2D, infos_gen_tmp->tex_part_1);

                // draw particles
                particules_1->afficher();

                // change the blending function back
                //glBlendFunc(GL_SRC_ALPHA,GL_DST_ALPHA);
		//glBlendFunc( GL_SRC_ALPHA,GL_ONE_MINUS_DST_ALPHA);
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
        }

}
